package Go;

import java.util.Enumeration;
import java.util.Vector;


// the current position in the game
public class GoPosition implements GoConstants, Cloneable {

   // all the points on the goban
  public GobanLocation[][] goban;

  // goban size
  private int size;
  
  //list of black & white groups
  private ListOfGroups blackGroups;
  private ListOfGroups whiteGroups;

  private Move lastMove;

  // number of captured Black stones
  private int nrBlackPrisoners = 0;

  // number of captured White stones
  private int nrWhitePrisoners = 0;

  // constructor for an empty position of "size" size goban
  public GoPosition (int size) {
    this.size = size;

    goban = new GobanLocation[size][size];

    blackGroups = new ListOfGroups();
    whiteGroups = new ListOfGroups();
    setGoban(size);

    lastMove = new Move(GoConstants.WHITE);
  }

  /*public Object clone() throws CloneNotSupportedException{
    GoPosition clonedGoPosition = new GoPosition(size);

    for (int i = 0; i < size; i++)
      for (int k = 0; k < size; k++)
        clonedGoPosition.goban[i][k] = this.goban[i][k];

    // todo: to be honest, we should also clone blackGroups
    // and whiteGroups lists... but for the moment we only
    // need to copy the "goban"

    return clonedGoPosition;

    // this is bad: the cloned GoPosition shares the same
    // reference to "goban" as "his" does...
    // return super.clone();
  } */

  public Object cloneCopy() throws CloneNotSupportedException{
    GobanLocation currentPoint;
    ListOfStones changedLocations = new ListOfStones();

    GoPosition clonedGoPosition = new GoPosition(size);

    int i, k;

    for (i = 0; i < size; i++)
      for (k = 0; k < size; k++) {
        currentPoint = clonedGoPosition.goban[i][k];
        currentPoint.state = this.goban[i][k].state;
        if (currentPoint.state != GoConstants.EMPTY)
          clonedGoPosition.setNewMove(currentPoint, changedLocations);
      }

    clonedGoPosition.nrBlackPrisoners = this.nrBlackPrisoners;
    clonedGoPosition.nrWhitePrisoners = this.nrWhitePrisoners;

    // I tried here to restore all the relations inside the cloned GoPosition,
    // but I encountered too many problems...
    /*for (i = 0; i < size; i++)
      for (k = 0; k < size; k++)
        clonedGoPosition.goban[i][k].state = this.goban[i][k].state;

    clonedGoPosition.setBlackGroups((ListOfGroups)(this.blackGroups.clone()));
    clonedGoPosition.setWhiteGroups((ListOfGroups)(this.whiteGroups.clone()));

    // VERY IMPORTANT: should restore all the relations inside the GoPosition,
    // like: goban[i][k].group !!!

    Group currentGroup;
    GobanLocation currentStone;
    Enumeration se;
    Enumeration ge;

    // get all black groups
    ge = clonedGoPosition.getBlackGroups().getElements();
    for (;ge.hasMoreElements();) {
      currentGroup = (Group)ge.nextElement();
      // get all stones
      se = currentGroup.stones.getElements();
      for (;se.hasMoreElements();) {
        currentStone = (GobanLocation)se.nextElement();
        currentStone = clonedGoPosition.goban[currentStone.x][currentStone.y];
        currentStone.setGroup(currentGroup);
      }
    }

    // get all white groups
    ge = clonedGoPosition.getWhiteGroups().getElements();
    for (;ge.hasMoreElements();) {
      currentGroup = (Group)ge.nextElement();
      // get all stones
      se = currentGroup.stones.getElements();
      for (;se.hasMoreElements();) {
        currentStone = (GobanLocation)se.nextElement();
        currentStone = clonedGoPosition.goban[currentStone.x][currentStone.y];
        currentStone.setGroup(currentGroup);
      }
    }

    clonedGoPosition.setLastMove(new Move(lastMove.x, lastMove.y, lastMove.color));*/

    return clonedGoPosition;
  }

  public void copyFrom(GoPosition source) {
    for (int i = 0; i < size; i++)
      for (int k = 0; k < size; k++)
        this.goban[i][k].state = source.goban[i][k].state;

    this.nrBlackPrisoners = source.nrBlackPrisoners;
    this.nrWhitePrisoners = source.nrWhitePrisoners;
  }

  // tells if a move is legal or not
  public boolean isLegalMove(Move lastMove) {
    int x, y, color;

    x = lastMove.x;
    y = lastMove.y;
    color = lastMove.color;

    if (goban[x][y].state != GoConstants.EMPTY)
      return false;

    GobanLocation lastMoveLocation = new GobanLocation(x, y, color, this);

    // if lastMove has at least one free liberty, it's legal
    if (countLiberties(x, y) > 0)
      return true;

    ListOfGroups listOfNeighborFriends = new ListOfGroups();
    ListOfGroups listOfNeighborEnemies = new ListOfGroups();

    // fill the lists with new stone's friend and enemy groups
    visitNewStoneNeighbors(lastMoveLocation,
                       listOfNeighborFriends,
                       listOfNeighborEnemies);

    // check if we would capture anything
    Enumeration enemies = listOfNeighborEnemies.getElements();
    for (;enemies.hasMoreElements();) {
      if (((Group)enemies.nextElement()).liberties.length() <= 1)
        return true;
    }

    // check if we commit suicide by this move
    int newOwnGroupNrLiberties = 0;

    Enumeration friends = listOfNeighborFriends.getElements();
    for (;friends.hasMoreElements();) {
      newOwnGroupNrLiberties += ((Group)friends.nextElement()).liberties.length();
      newOwnGroupNrLiberties -= 1;
    }

    if (newOwnGroupNrLiberties > 0) return true;
    else return false;
  }

  public boolean isSelfAtariMove(Move lastMove){
    int x, y, color;

    x = lastMove.x;
    y = lastMove.y;
    color = lastMove.color;

    GobanLocation lastMoveLocation = new GobanLocation(x, y, color, this);

    // if lastMove has at least one free liberty, it's legal
    if (countLiberties(x, y) > 1)
      return false;

    ListOfGroups listOfNeighborFriends = new ListOfGroups();
    ListOfGroups listOfNeighborEnemies = new ListOfGroups();

    // fill the lists with new stone's friend and enemy groups
    visitNewStoneNeighbors(lastMoveLocation,
                       listOfNeighborFriends,
                       listOfNeighborEnemies);

    // check if we would capture anything
    Enumeration enemies = listOfNeighborEnemies.getElements();
    for (;enemies.hasMoreElements();) {
      if (((Group)enemies.nextElement()).liberties.length() <= 1)
        return false;
    }

    // check if we commit suicide by this move
    int newOwnGroupNrLiberties = 0;

    Enumeration friends = listOfNeighborFriends.getElements();
    for (;friends.hasMoreElements();) {
      newOwnGroupNrLiberties += ((Group)friends.nextElement()).liberties.length();
      newOwnGroupNrLiberties -= 1;
    }

    if (newOwnGroupNrLiberties > 1) return false;
    else return true;
  }

  // returns the Location corresponding to a Move
  public GobanLocation newMove(Move theNewMove) {
    goban[theNewMove.x][theNewMove.y].state = theNewMove.color;
    return goban[theNewMove.x][theNewMove.y];
  }

  // a new move is added; the corresponding GobanLocation is returned
  public GobanLocation addNewStone(int x, int y, int color) {
    goban[x][y].state = color;
    return goban[x][y];
  }

  public void setSize(int Size) {
    size = Size;
  }

  public int getSize() {
    return size;
  }

  //a new stone has been placed on the board
  public void setNewMove(GobanLocation stone,
                         ListOfStones changedLocations) {

    // build helping lists
    ListOfGroups listOfNeighborFriends = new ListOfGroups();

    ListOfGroups listOfNeighborEnemies = new ListOfGroups();

    // fill the lists with new stone's friend and enemy groups
    visitNewStoneNeighbors(stone,
                       listOfNeighborFriends,
                       listOfNeighborEnemies);

    //once we placed a move, it's location is no longer
    // a liberty for the surrounding friend and enemy groups
    removeNeighborsLiberty(stone,
                           listOfNeighborFriends,
                           listOfNeighborEnemies);

    // update the situation of the friend groups (listOfNeighborFriends) after the last move (stone)
    updateStonesFriendGroups(stone, listOfNeighborFriends);

    // take care of captured stones
    updateStonesEnemyGroups(stone,
                            listOfNeighborEnemies,
                            changedLocations);
    setLastMove(new Move(stone.x, stone.y, stone.state));
  } // setNewMove

  // returns a new GoPosition, resulted after playing <stone>
  public GoPosition positionAfterMove(Move stone) {
    GoPosition newPosition = null;
    try {
      newPosition = (GoPosition)this.cloneCopy();

      newPosition.setLastMove(stone);

      // the position will not be modified in case of a PASS move!
      if (!stone.isPass()) {
        ListOfStones changedLocations = new ListOfStones();

        newPosition.goban[stone.x][stone.y].state = stone.color;
        newPosition.setNewMove(newPosition.goban[stone.x][stone.y],
                               changedLocations);
      }
    }
    catch (CloneNotSupportedException ex) {
      System.out.println("Problems in GoPosition.positionAfterMove: " + ex);
      System.exit(1);
    }

    return newPosition;
  }
  
  public Enumeration legalMoves() {
    Vector moves = new Vector();
    Move currentMove;
    int currentColor = turn();

    for (int i = 0; i < size; i++)
      for (int j = 0; j < size; j++) {
        currentMove = new Move(i, j, currentColor);
        if (isLegalMove(currentMove))
          moves.addElement(currentMove);
      }

    // a PASS move is always legal ... :-)
    //moves.addElement(new Move(GoConstants.PASS,
      //                        GoConstants.PASS,
        //                      currentColor));

    return moves.elements();
  }

  // same as legalMoves, but returns GobanLocations, not Moves
  public Enumeration legalMoves_GL() {
    Vector moves = new Vector();
    Move currentMove;
    int currentColor = turn();

    for (int i = 0; i < size; i++)
      for (int j = 0; j < size; j++) {
        currentMove = new Move(i, j, currentColor);
        if (isLegalMove(currentMove))
          moves.addElement(goban[i][j]);
      }

    // a PASS move is always legal ... :-)
    //moves.addElement(new Move(GoConstants.PASS,
      //                        GoConstants.PASS,
        //                      currentColor));

    return moves.elements();
  }

  public ListOfGroups getBlackGroups() {
    return blackGroups;
  }

  public ListOfGroups getWhiteGroups() {
    return whiteGroups;
  }

  public void setBlackGroups(ListOfGroups newBlackGroups) {
    blackGroups = newBlackGroups;
  }

  public void setWhiteGroups(ListOfGroups newWhiteGroups) {
    whiteGroups = newWhiteGroups;
  }


  public boolean isEmpty(int x, int y) {
    return (goban[x][y].state == EMPTY);
  }

  public String toString() {
    return "{B=" + blackGroups + "; W=" + whiteGroups + "}";
  }

  // fast method to count liberties for the (x, y) goban location
  public int countLiberties(int x, int y) {
    int count = 0;

    if (x > 0)
      if (goban[x-1][y].state == EMPTY) count++;

    if (x < size - 1)
      if (goban[x+1][y].state == EMPTY) count++;

    if (y > 0)
      if (goban[x][y-1].state == EMPTY) count++;

    if (y < size - 1)
      if (goban[x][y+1].state == EMPTY) count++;

    return count;  
  }

  public Move getLastMove() {
    return lastMove;
  }

  public void setLastMove(Move move){
    lastMove.x = move.x;
    lastMove.y = move.y;
    lastMove.color = move.color;
  }

  public int turn() {
    if (lastMove.color == GoConstants.BLACK) return GoConstants.WHITE;
    else return GoConstants.BLACK;
  }

  public int getNrBlackPrisoners() { return nrBlackPrisoners; }

  public int getNrWhitePrisoners() { return nrWhitePrisoners; }

  // todo !!! improve this, is too slow for big boards
  // compare two board positions
  public boolean sameAs(GoPosition somePosition) {
    for (int i = 0; i < size; i++)
      for (int k = 0; k < size; k++)
        if (goban[i][k].state != somePosition.goban[i][k].state)
          return false;

    return true;
  }

  private void setGoban(int Size) {
    size = Size;
    //goban = new GobanLocation[size][size];
    for (int i = 0; i < size; i++)
       for (int k = 0; k < size; k++) {
          goban[i][k] = new GobanLocation(i, k, EMPTY, this);
       }

    for (int i = 0; i < size; i++)
       for (int k = 0; k < size; k++) {
          goban[i][k].computeLiberties();
       }

  } // set Goban

  // implementation procedure used in setNextMove
  // process the "stone"'s neighbor location ("currentX","currentY")
  private void processNeighbor(GobanLocation stone,
                              int currentX, int currentY,
                              ListOfGroups listOfNeighborFriends,
                              ListOfGroups listOfNeighborEnemies) {

   // the currrently searched location is empty
   if (goban[currentX][currentY].state == EMPTY) {
     stone.liberties.add(goban[currentX][currentY]);
     goban[currentX][currentY].liberties.deleteElement(stone);
     return;
   }

   // the currrently searched location is occupied by a friend;
   // include that stone's group in the listOfNeighborFriends
   else if (goban[currentX][currentY].state == stone.state) {
      listOfNeighborFriends.add(goban[currentX][currentY].getGroup());
   }
         // the currrently searched location is occupied by an enemy
   // include that stone's group in the listOfNeighborEnemies
   else {
      listOfNeighborEnemies.add(goban[currentX][currentY].getGroup());
   }
  } // processNeighbor


  //once we placed a move, it's location is no longer
  // a liberty for the surrounding friend stones
  private void removeNeighborsLiberty(GobanLocation stone,
                                ListOfGroups listOfNeighborFriends,
                                ListOfGroups listOfNeighborEnemies) {

   Enumeration e = listOfNeighborFriends.getElements();
   for (;e.hasMoreElements();) {
     Group currentGroup = (Group)e.nextElement();
     currentGroup.liberties.deleteElement(stone);
   }

   //once we placed a move, it's location is no longer
   // a liberty for the surrounding enemy stones
   e = listOfNeighborEnemies.getElements();
   for (;e.hasMoreElements();) {
     Group currentGroup = (Group)e.nextElement();
     currentGroup.liberties.deleteElement(stone);
   }
  } //removeNeighborsLiberty



  // update the situation of the friend groups (listOfNeighborFriends) after the last move (stone)
  private void updateStonesFriendGroups(GobanLocation stone,
                                       ListOfGroups listOfNeighborFriends) {
   // no friends around...means a new group has been born!
   if (listOfNeighborFriends.length() == 0)
     // build a new group that contains only one stone
     newBornGroup(stone);

   //or exactly one friend...
   else if (listOfNeighborFriends.length() == 1)
     // the last move (stone) should be appended to the only neighbor friend group
     addStoneToGroup(stone,
                     listOfNeighborFriends.first());

   //or more than one friend
   else
     // the last move (stone) is connecting several friend groups (listOfNeighborFriends)
     connectMoreGroups(stone,
                       listOfNeighborFriends);
  } //updateStonesFriendGroups



  // update the situation of the enemy groups (listOfNeighborEnemies) after the last move (stone)
  private void updateStonesEnemyGroups(GobanLocation stone,
                                       ListOfGroups listOfNeighborEnemies,
                                       ListOfStones changedLocations) {

   Enumeration e = listOfNeighborEnemies.getElements();
   for(;e.hasMoreElements();) {
     Group currentGroup = (Group)e.nextElement();
     if (currentGroup.hasNoLiberty()) {
       // update the captured stones info
       if (currentGroup.color == GoConstants.BLACK) { nrBlackPrisoners += currentGroup.nrStones(); }
       else { nrWhitePrisoners += currentGroup.nrStones(); }

       // add the stones of this group to the list of changed locations
       changedLocations.appendList(currentGroup.stones);

       // iterate for all the stones in the "to be captured" enemy group
       Enumeration stonesList = currentGroup.stones.getElements();
       for(;stonesList.hasMoreElements();) {
         GobanLocation currentStone = (GobanLocation)stonesList.nextElement();

         // update the liberties after removing currentStone
         Enumeration enemyList = (currentStone.getNeighborEnemies()).getElements();
         for (;enemyList.hasMoreElements();) {
           Group currentEnemy = (Group)enemyList.nextElement();
           currentEnemy.liberties.add(currentStone);
         }

         // we actually remove the currentStone here
         currentStone.state = EMPTY;

         if (currentGroup.color == BLACK)
           blackGroups.deleteElement(currentGroup);
         else
         if (currentGroup.color == WHITE)
           whiteGroups.deleteElement(currentGroup);
       } // for
     } // if
   } // for
  } //updateStonesEnemyGroups


  // fill the lists with stone's friend and enemy groups
  public void visitNewStoneNeighbors(GobanLocation stone,
                        ListOfGroups listOfNeighborFriends,
                        ListOfGroups listOfNeighborEnemies) {

   // reset the stone's liberties info: it will be built within this method
   /*if (stone.liberties != null) {*/
     //stone.liberties.removeAllElements();
  /* }
   else
     stone.liberties = new ListOfLiberties();*/

   if (stone.x > 0) {
       int currentX = stone.x - 1;
       int currentY = stone.y;

       processNeighbor(stone,
                       currentX, currentY,
                       listOfNeighborFriends,
                       listOfNeighborEnemies);
    }

    if (stone.y > 0) {
       int currentX = stone.x;
       int currentY = stone.y - 1;

       processNeighbor(stone,
                       currentX, currentY,
                       listOfNeighborFriends,
                       listOfNeighborEnemies);
    }

    if (stone.x < size - 1) {
       int currentX = stone.x + 1;
       int currentY = stone.y;

       processNeighbor(stone,
                       currentX, currentY,
                       listOfNeighborFriends,
                       listOfNeighborEnemies);
    }

    if (stone.y < size - 1) {
       int currentX = stone.x;
       int currentY = stone.y + 1;

       processNeighbor(stone,
                       currentX, currentY,
                       listOfNeighborFriends,
                       listOfNeighborEnemies);
    }
  } // visitStoneNeighbors


  private void newBornGroup(GobanLocation stone) {
   Group newGroup = new Group(stone);
   stone.setGroup(newGroup);

   //add this group to the appropriate list of groups
   //System.out.println("[GoPosition] Add the new group to the list of groups");
   if (newGroup.color == BLACK) {
     //System.out.println("It's the Black group " + newGroup + " to add to list " + blackGroups);
     //System.out.println("Position is now " + this + " and blackGroupList: " + this.blackGroups);
     blackGroups.add(newGroup);
   }
   else {
     //System.out.println("It's the White group " + newGroup + " to add to list " + whiteGroups);
      whiteGroups.add(newGroup);
   }
  } //newBornGroup


  // adding a stone to the group (theGroup)
  private void addStoneToGroup(GobanLocation stone,
                              Group theGroup) {
   //System.out.println("[addStoneToGroup] We have one friend: the group " + theGroup);

   // add the stone to the group's list
   theGroup.stones.add(stone);

   //System.out.println("Setting the stone's 'group' property...");
   stone.setGroup(theGroup);

   theGroup.liberties.appendList(stone.liberties);
  } // addStoneToOneGroup



  // the last move (stone) is connecting several friend groups (listOfNeighborFriends)
  private void connectMoreGroups(GobanLocation stone,
                                ListOfGroups listOfNeighborFriends) {
   //System.out.println("[connectMoreGroups]We have more friends:" + listOfNeighborFriends);
   // get a handle of the first friend group in the list
   Group theGroup = listOfNeighborFriends.first();

   //theGroup.liberties.appendList(stone.liberties);

   // the other friend groups in the list will be embeded in the first one
   Enumeration e = listOfNeighborFriends.getElements();
   e.nextElement();

   //System.out.println("[connectMoreGroups]entered for loop");
   for (;e.hasMoreElements();) {


     // get a handle of the next friend group in the list
     Group oneMoreGroup = (Group)e.nextElement();

     // chsnge the apartenence of the stones in the group to be "swalowed"
     oneMoreGroup.stones.setGroup(theGroup);

     // swallow the liberties and the stones of one group into the other
     theGroup.liberties.appendList(oneMoreGroup.liberties);
     theGroup.stones.appendList(oneMoreGroup.stones);

     // delete the second group from the coresponding group list
     if (stone.state == BLACK)
       blackGroups.deleteElement(oneMoreGroup);
     else
       whiteGroups.deleteElement(oneMoreGroup);
   } //for
   //System.out.println("[connectMoreGroups]exited for loop");

   // add the last move (stone) to the group, too
   addStoneToGroup(stone, theGroup);
  } //connectMoreGroups

}