package Go.strategy;

import Go.*;
import java.util.Enumeration;
import java.util.Vector;

public class BasicAtariGoStrategy {

  public static Move tryBasicMove(GoPosition goPosition) {
    Move theMove = null;

    if ((theMove = tryCapture(goPosition)) != null)
      return theMove;
    else if ((theMove = tryDefend(goPosition)) != null)
      return theMove;

    return theMove;      
  }

  // try to capture an enemy group
  public static Move tryCapture(GoPosition goPosition) {
    ListOfGroups listOfNeighborEnemies;
    ListOfLiberties libertiesList;

    if (goPosition.turn() == GoConstants.BLACK)
      listOfNeighborEnemies = goPosition.getWhiteGroups();
    else
      listOfNeighborEnemies = goPosition.getBlackGroups();

    Enumeration enemies = listOfNeighborEnemies.getElements();
    for (;enemies.hasMoreElements();) {
      libertiesList = ((Group)enemies.nextElement()).liberties;
      if (libertiesList.length() == 1) {
        GobanLocation location = libertiesList.getElementAt(0);
        return new Move(location.x, location.y, goPosition.turn());
      }
    }

    return null;
  }

  // try to defend own's groups against capture
  public static Move tryDefend(GoPosition goPosition) {
    ListOfGroups listOfNeighborFriends = new ListOfGroups();
    ListOfLiberties libertiesList;
    Move theMove;

    if (goPosition.turn() == GoConstants.WHITE)
      listOfNeighborFriends = goPosition.getWhiteGroups();
    else
      listOfNeighborFriends = goPosition.getBlackGroups();

    Enumeration friends = listOfNeighborFriends.getElements();
    for (;friends.hasMoreElements();) {
      libertiesList = ((Group)friends.nextElement()).liberties;
      if (libertiesList.length() == 1) {
        GobanLocation location = libertiesList.getElementAt(0);
        theMove = new Move(location.x, location.y, goPosition.turn());

        if (goPosition.isLegalMove(theMove))
          return theMove;
      }
    }
    return null;
  }

  // returns an enumeration with the possible contact moves
  public static Enumeration contactMoves(GoPosition goPosition) {
    Enumeration blackGroups, whiteGroups;
    ListOfLiberties result, currentLiberties;
    GobanLocation currentLiberty;

    result = new ListOfLiberties();

    whiteGroups = goPosition.getWhiteGroups().getElements();
    blackGroups = goPosition.getBlackGroups().getElements();

    // add black groups' liberties to the list
    for (;blackGroups.hasMoreElements();) {
      // get current group's liberties
      result.appendList(((Group)blackGroups.nextElement()).liberties);
    }

    // add white groups' liberties to the list
    for (;whiteGroups.hasMoreElements();) {
      // get current group's liberties
      result.appendList(((Group)whiteGroups.nextElement()).liberties);
    }

    //return filterLegalMoves(goPosition, result).getElements();
    return filterSelfAtari(goPosition, filterLegalMoves(goPosition, result)).getElements();
  }

  // return an Enumeration of the liberties for weakest enemy group and
  // weakest own group
  public static Enumeration extremeAttackDefenseMoves(GoPosition goPosition) {
    ListOfLiberties result = new ListOfLiberties();
    Move theMove, bestMove = null;

    ListOfGroups listOfEnemies;
    ListOfGroups listOfFriends;

    ListOfLiberties libertiesList;
    Group currentWeakest = null, currentGroup = null;
    int currentNrLiberties = 1000, bestNrliberties = -1;

    if (goPosition.turn() == GoConstants.BLACK) {
      listOfEnemies = goPosition.getWhiteGroups();
      listOfFriends = goPosition.getBlackGroups();
    }
    else {
      listOfEnemies = goPosition.getBlackGroups();
      listOfFriends = goPosition.getWhiteGroups();
    }

    Enumeration enemies = listOfEnemies.getElements();
    Enumeration friends = listOfFriends.getElements();

    // find weakest enemy group
    for (;enemies.hasMoreElements();) {
      currentGroup = (Group)enemies.nextElement();
      libertiesList = currentGroup.liberties;

      if (libertiesList.length() < currentNrLiberties) {
        currentNrLiberties = libertiesList.length();
        currentWeakest = currentGroup;
      }
    }

    if (currentWeakest != null)
      // append opponent's weakeast group's liberties
      result.appendList(currentWeakest.liberties);

    currentWeakest = null;

    // find weakest own group
    for (;friends.hasMoreElements();) {
      currentGroup = (Group)friends.nextElement();
      libertiesList = currentGroup.liberties;

      if (libertiesList.length() < currentNrLiberties) {
        currentNrLiberties = libertiesList.length();
        currentWeakest = currentGroup;
      }
    }

    if (currentWeakest != null)
      // append opponent's weakeast group's liberties
      result.appendList(currentWeakest.liberties);

    return filterSelfAtari(goPosition, filterLegalMoves(goPosition, result)).getElements();
  }

  // it filters the <list> of Moves (second argument), returning only those Moves
  // which are legal in the given <goPosition>
  private static ListOfLiberties filterLegalMoves(GoPosition goPosition,
                                         ListOfLiberties list) {
    // we have a list  of liberties; should get an Enumeration with the
    // corresponding Moves instead, and only the legal ones!
    ListOfLiberties moves = new ListOfLiberties();

    int nextColor = goPosition.turn();

    Enumeration e = list.getElements();
    for (;e.hasMoreElements();) {
      GobanLocation location = (GobanLocation)e.nextElement();
      Move currentMove = new Move(location.x, location.y, nextColor);

      // only add the legal moves in the list
      if (goPosition.isLegalMove(currentMove))
        moves.add(new GobanLocation(location));
    }

    return moves;
  }

  // it filters the <list> of Moves (second argument), returning only those Moves
  // which are not placing the Player to play it in Atari (i.e. worst possible move in an atari-go game)
  private static ListOfLiberties filterSelfAtari(GoPosition goPosition,
                                      ListOfLiberties list) {
    // we have a list  of liberties; should get an Enumeration with the
    // corresponding Moves instead, and only the legal ones!
    ListOfLiberties moves = new ListOfLiberties();

    int nextColor = goPosition.turn();

    Enumeration e = list.getElements();
    for (;e.hasMoreElements();) {
      GobanLocation location = (GobanLocation)e.nextElement();
      Move currentMove = new Move(location.x, location.y, nextColor);

      // only add the legal moves in the list
      if (! goPosition.isSelfAtariMove(currentMove))
        moves.add(new GobanLocation(location));
    }

    return moves;
  }

}
